package de.bsvrz.sys.usv.cs121;

import java.io.IOException;

import org.snmp4j.CommandResponder;
import org.snmp4j.CommandResponderEvent;
import org.snmp4j.CommunityTarget;
import org.snmp4j.PDU;
import org.snmp4j.Snmp;
import org.snmp4j.TransportMapping;
import org.snmp4j.log.JavaLogFactory;
import org.snmp4j.log.LogFactory;
import org.snmp4j.mp.SnmpConstants;
import org.snmp4j.smi.Address;
import org.snmp4j.smi.GenericAddress;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.OctetString;
import org.snmp4j.smi.VariableBinding;
import org.snmp4j.transport.DefaultUdpTransportMapping;

import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.usv.enums.BatterieZustand;
import de.bsvrz.sys.usv.enums.EingangsfrequenzZustand;
import de.bsvrz.sys.usv.enums.KommunikationsZustand;
import de.bsvrz.sys.usv.enums.USVZustand;
import de.bsvrz.sys.usv.enums.WechselSpannung;
import de.bsvrz.sys.usv.enums.ZustandNotstrom;
import de.bsvrz.sys.usv.enums.ZustandPhase;
import de.bsvrz.sys.usv.usv.DefaultUSV;
import de.bsvrz.sys.usv.wrapper.USVAllgemein;
import de.bsvrz.sys.usv.wrapper.USVWrapper;

/**
 * Service-Stub für die Implementierung eines konkreten USV-Typs. Hierbei
 * handelt es sich um die USV der Firma Wöhrle (www.woehrle-svs.de), die in der
 * VRZ Stuttgart eingesetzt wird.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann / Christian Hösel
 */
public class UsvCS121 extends DefaultUSV {

	static {
		LogFactory.setLogFactory(new JavaLogFactory());
	}

	/**
	 * RFC 1628: The MIB module to describe Uninterruptible Power Supplies.
	 */
	public static final String UPS_MIB = "1.3.6.1.2.1.33";

	/**
	 * RFC 1628: An estimate of the battery charge remaining expressed as a
	 * percent of full charge.
	 *
	 * <p>
	 * Wertebereich laut RFC 1628: 0..100
	 */
	public static final String UPS_ESTIMATED_CHARGE_REMAINING = UsvCS121.UPS_MIB + ".1.2.4.0";

	/**
	 * RFC 1628: The magnitude of the present input voltage.
	 *
	 * <p>
	 * Wertebereich: INTEGER (0..2147483647)
	 */
	public static final String UPS_INPUT_VOLTAGE = UsvCS121.UPS_MIB + ".1.3.3.1.3";

	/**
	 * The name of the UPS manufacturer.
	 */
	public static final String UPS_IDENT_MANUFACTURER = UsvCS121.UPS_MIB + ".1.1.1.0";

	/**
	 * The present input frequency.
	 *
	 * <p>
	 * Wertebereich: INTEGER (0..2147483647)
	 */
	public final static String UPS_INPUT_FREQUENCY = UsvCS121.UPS_MIB + ".1.3.3.1.2";

	/**
	 * If the unit is on battery power, the elapsed time since the UPS last
	 * switched to battery power, or the time since the network management
	 * subsystem was last restarted, whichever is less. Zero shall be returned
	 * if the unit is not on battery power.
	 */
	public final static String UPS_SECONDS_ON_BATTERY = UsvCS121.UPS_MIB + ".1.2.2.0";

	/**
	 * The magnitude of the present battery voltage.
	 */
	public final static String UPS_BATTERY_VOLTAGE = UsvCS121.UPS_MIB + ".1.2.5.0";

	/**
	 * The ambient temperature at or near the UPS Battery casing.
	 */
	public final static String UPS_BATTERY_TEMPERATURE = UsvCS121.UPS_MIB + ".1.2.7.0";

	/**
	 * An estimate of the time to battery charge depletion under the present
	 * load conditions if the utility power is off and remains off, or if it
	 * were to be lost and remain off.
	 */
	public final static String UPS_ESTIMATED_MINUTES_REMAINING = UsvCS121.UPS_MIB + ".1.2.3.0";

	/**
	 * RFC 1628: The present source of output power. The enumeration none(2)
	 * indicates that there is no source of output power (and therefore no
	 * output power), for example, the system has opened the output breaker.
	 *
	 * <p>
	 * Mögliche Zustände laut RFC 1628:
	 * <ul>
	 * <li>other(1),</li>
	 * <li>none(2),</li>
	 * <li>normal(3),</li>
	 * <li>bypass(4),</li>
	 * <li>battery(5),</li>
	 * <li>booster(6),</li>
	 * <li>reducer(7)</li>
	 * </ul>
	 */
	public static final String UPS_OUTPUT_SOURCE = UsvCS121.UPS_MIB + ".1.4.1";
	private static int UPS_OUTPUT_SOURCE_NORMAL = 3;
	private static int UPS_OUTPUT_SOURCE_BATTERY = 5;

	private final Debug log = Debug.getLogger();

	private USVWrapper usvWrapper;

	/**
	 * Der Timeout (in Millisekunden), der für alle Anfragen an der USV
	 * verwendet wird. Default ist {@link USVAllgemein#DEFAULT_TIMEOUT}.
	 */
	private long timeout = USVAllgemein.DEFAULT_TIMEOUT;

	/**
	 * Default-Adresse
	 */
	private String usvAddress = "127.0.0.1/4700";

	@Override
	public int getBatterieZustand() {
		int result = BatterieZustand.na.getValue();
		log.fine("Starte Abfrage des Batteriezustandes (OID=" + UsvCS121.UPS_ESTIMATED_CHARGE_REMAINING + ") ...");
		final PDU response = sendRequest(UsvCS121.UPS_ESTIMATED_CHARGE_REMAINING);
		try {
			if (response != null) {
				result = response.get(0).getVariable().toInt();
				log.info("Ergebnis Batteriezustand: " + result + " %");
			}
		} catch (final Exception e) {
			log.warning(
					"Die Auswertung des Ergebnisses der Abfrage des Batteriezustandes hat zu einer Ausnahmen geführt. ",
					e);
		}
		return result;
	}

	@Override
	public EingangsfrequenzZustand getEingangsFrequenzZustand() {
		EingangsfrequenzZustand result = EingangsfrequenzZustand.na;

		// Maximale Frequenz --> 105% von Normal
		final float maxFrequenz = 52.5f;
		// Minimale Frequenz --> 95% von Normal
		final float minFrequenz = 47.5f;

		try {
			// An dieser Stelle gibt es 3 zu ermittelnde Frequenzen, die sich
			// in einer Tabelle befinden.
			for (int i = 1; i <= 3; i++) {
				log.fine("Starte Abfrage der Eingangsfrequenz L" + i + " (OID=" + UsvCS121.UPS_INPUT_VOLTAGE + "." + i
						+ ") ...");
				final PDU response = sendRequest(UsvCS121.UPS_INPUT_FREQUENCY + "." + i);
				if (response != null) {
					final float frequenz = response.get(0).getVariable().toInt() / 10f;
					// Testen gegen Grenzen und Ermittlung des Zustandes
					if (frequenz > maxFrequenz) {
						result = EingangsfrequenzZustand.hoch;
					} else if (frequenz < minFrequenz) {
						result = EingangsfrequenzZustand.niedrig;
					} else {
						result = EingangsfrequenzZustand.normal;

					}
					log.info("Ergebnis Eingangsfrequenz L" + i + ": " + frequenz + " Hz--> Zustand: "
							+ result.toString());
					if (result != EingangsfrequenzZustand.normal) {
						break;
					}
				}
			}
		} catch (final Exception e) {
			result = EingangsfrequenzZustand.na;
			log.warning(
					"Die Auswertung des Ergebnisses der Abfrage der Eingangsfrequenz hat zu einer Ausnahmen geführt. ",
					e);
		}
		return result;
	}

	@Override
	public KommunikationsZustand getKommunikationsZustand() {
		KommunikationsZustand result = KommunikationsZustand.Störung;
		log.fine("Starte Kommunikationstest (OID=" + UsvCS121.UPS_IDENT_MANUFACTURER + ") ...");
		final PDU response = sendRequest(UsvCS121.UPS_IDENT_MANUFACTURER);
		if (response != null) {
			result = KommunikationsZustand.normal;
			log.info("Ergebnis Kommunikationstest: " + result);
		}
		return result;
	}

	@Override
	public USVZustand getUsvZustand() {
		USVZustand result = USVZustand.na;
		log.fine("Starte Abfrage des USV-Zustandes (OID=" + UsvCS121.UPS_OUTPUT_SOURCE + ") ...");
		final PDU response = sendRequest(UsvCS121.UPS_OUTPUT_SOURCE);
		if (response != null) {
			try {
				final Integer outputSource = response.get(0).getVariable().toInt();

				if (outputSource == UsvCS121.UPS_OUTPUT_SOURCE_NORMAL) {
					result = USVZustand.normal;
				} else if (outputSource == UsvCS121.UPS_OUTPUT_SOURCE_BATTERY) {
					result = USVZustand.Batterie;

				}
			} catch (final Exception e) {
				log.warning(
						"Die Auswertung des Ergebnisses der Abfrage des USV-Zustandes hat zu einer Ausnahmen geführt. ",
						e);
			}
			log.info("Ergebnis USV-Zustand: " + result.toString());
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see de.bsvrz.sys.usv.usv.DefaultUSV#getBatterieSpannung()
	 */
	@Override
	public int getBatterieSpannung() {
		int result = -1;
		log.fine("Starte Abfrage des USV-Batteriespannung (OID=" + UsvCS121.UPS_BATTERY_VOLTAGE + ") ...");
		final PDU response = sendRequest(UsvCS121.UPS_BATTERY_VOLTAGE);
		if (response != null) {
			try {
				result = response.get(0).getVariable().toInt();

			} catch (final Exception e) {
				log.warning(
						"Die Auswertung des Ergebnisses der Abfrage der USV-Batteriespannung hat zu einer Ausnahmen geführt. ",
						e);
			}
			log.info("Ergebnis USV-Batteriespannung: " + result + " V.");
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see de.bsvrz.sys.usv.usv.DefaultUSV#getBatterieTemperatur()
	 */
	@Override
	public int getBatterieTemperatur() {
		int result = -1;
		log.fine("Starte Abfrage des USV-Batterietemperatur (OID=" + UsvCS121.UPS_BATTERY_TEMPERATURE + ") ...");
		final PDU response = sendRequest(UsvCS121.UPS_BATTERY_TEMPERATURE);
		if (response != null) {
			try {
				result = response.get(0).getVariable().toInt();

			} catch (final Exception e) {
				log.warning(
						"Die Auswertung des Ergebnisses der Abfrage der USV-Batterietemperatur hat zu einer Ausnahmen geführt. ",
						e);
			}
			log.info("Ergebnis USV-Batterietemperatur: " + result + " °C.");
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see de.bsvrz.sys.usv.usv.DefaultUSV#getBetriebsdauerBatterie()
	 */
	@Override
	public int getBetriebsdauerBatterie() {
		int result = -1;
		log.fine("Starte Abfrage des USV-Batteriebetriebsdauer (OID=" + UsvCS121.UPS_SECONDS_ON_BATTERY + ") ...");
		final PDU response = sendRequest(UsvCS121.UPS_SECONDS_ON_BATTERY);
		if (response != null) {
			try {
				result = response.get(0).getVariable().toInt();

			} catch (final Exception e) {
				log.warning(
						"Die Auswertung des Ergebnisses der Abfrage der USV-Batteriebetriebsdauer hat zu einer Ausnahmen geführt. ",
						e);
			}
			log.info("Ergebnis USV-Batteriebetriebsdauer: " + result + " Sekunden.");
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see de.bsvrz.sys.usv.usv.DefaultUSV#getUSVRestlaufzeit()
	 */
	@Override
	public int getUSVRestlaufzeit() {
		int result = -1;
		log.fine("Starte Abfrage des USV-Batterierestlaufzeit (OID=" + UsvCS121.UPS_ESTIMATED_MINUTES_REMAINING
				+ ") ...");
		final PDU response = sendRequest(UsvCS121.UPS_ESTIMATED_MINUTES_REMAINING);
		if (response != null) {
			try {
				result = response.get(0).getVariable().toInt();

			} catch (final Exception e) {
				log.warning(
						"Die Auswertung des Ergebnisses der Abfrage der USV-Batterierestlaufzeit hat zu einer Ausnahmen geführt. ",
						e);
			}
			log.info("Ergebnis USV-Batterierestlaufzeit: " + result + " Minuten.");
		}
		return result;
	}

	@Override
	public WechselSpannung getWechselSpannung() {
		WechselSpannung result = WechselSpannung.na;

		// Entspricht 115% von 230V
		final float maxSpannungOK = 264.5f;

		// Entspricht 75% von 230V
		final float minSpannungOK = 172.5f;

		try {
			// An dieser Stelle gibt es 3 zu ermittelnde Wechselspannungen, die
			// sich
			// in einer Tabelle befinden.
			for (int i = 1; i <= 3; i++) {
				log.fine("Starte Abfrage der WechselSpannung L" + i + " (OID=" + UsvCS121.UPS_INPUT_VOLTAGE + "." + i
						+ ") ...");
				final PDU response = sendRequest(UsvCS121.UPS_INPUT_VOLTAGE + "." + i);
				if (response != null) {
					final int spannung = response.get(0).getVariable().toInt();
					// Testen gegen Grenzen und Ermittlung des Zustandes
					if (spannung > maxSpannungOK) {
						result = WechselSpannung.Ueberspannung;
					} else if (spannung < minSpannungOK) {
						result = WechselSpannung.Unterspannung;
					} else {
						result = WechselSpannung.normal;
					}
					log.info("Ergebnis WechselSpannung L" + i + ": " + spannung + " V --> Zustand: "
							+ result.toString());
					if (result != WechselSpannung.normal) {
						break;
					}

				}
			}
		} catch (final Exception e) {
			result = WechselSpannung.na;
			log.warning(
					"Die Auswertung des Ergebnisses der Abfrage des Wechselspannungsfrequenz am Eingang hat zu einer Ausnahmen geführt. ",
					e);
		}
		return result;
	}

	/**
	 * Die Defaultimplementierung gibt immer "nicht ermittelbar" zurück.
	 */
	@Override
	public ZustandNotstrom getZustandNotstrom() {
		return ZustandNotstrom.na;
	}

	/**
	 * Die Defaultimplementierung gibt immer "nicht ermittelbar" zurück.
	 */
	@Override
	public ZustandPhase getZustandPhase1() {
		return ZustandPhase.na;
	}

	/**
	 * Die Defaultimplementierung gibt immer "nicht ermittelbar" zurück.
	 */
	@Override
	public ZustandPhase getZustandPhase2() {
		return ZustandPhase.na;
	}

	/**
	 * Die Defaultimplementierung gibt immer "nicht ermittelbar" zurück.
	 */
	@Override
	public ZustandPhase getZustandPhase3() {
		return ZustandPhase.na;
	}

	/**
	 * Gibt den USV-Wrapper zurück, der mit {@link #setInstance(USVWrapper)}
	 * zugewiesen wurde.
	 *
	 * @return der USV-Wrapper oder {@code null}, wenn keiner zugewiesen wurde.
	 * @see #setInstance(USVWrapper)
	 */
	protected USVWrapper getUsvWrapper() {
		return usvWrapper;
	}

	@Override
	public void setInstance(final USVWrapper instance) {
		usvWrapper = instance;
		timeout = usvWrapper.getUsvTimeout();
		log.info("Timeout für das Abrufen der USV Daten ist " + timeout + " Millisekunden.");

		// Auslesen der Adresse der USV aus dem DAV
		// final String pidUSV = USVWrapper.getPidUsv();
		// final SystemObject usvObj = USVWrapper.getConnection().getDataModel()
		// .getObject(pidUSV);
		// final AttributeGroup atgEigenschaften = USVWrapper.getConnection()
		// .getDataModel().getAttributeGroup("atg.usv");
		// usvAddress = usvObj.getConfigurationData(atgEigenschaften)
		// .getTextValue("USV-Typ").getText();

		// Auslesen der Adresse der USV aus einem Properties-File
		// ("usv.properties")
		// String fileLocation = ".." + File.separatorChar +
		// "distributionspakete"
		// + File.separatorChar + "de.bsvrz.dav.usv.cs121"
		// + File.separatorChar + "usv.properties";

		// /**
		// * Die Properties-Datei muss in dem Verzeichnis liegen, aus dem das
		// USV
		// * Modul gestartet wird.
		// */
		// String fileLocation = "usv.properties";
		// final String addressPropertyKey = "usv_adresse";
		// final Properties p = new Properties();
		// final File propertiesFile = new File(fileLocation);
		// try {
		// fileLocation = propertiesFile.getCanonicalPath();
		// if (propertiesFile.exists() && propertiesFile.canRead()) {
		// p.load(new FileInputStream(propertiesFile));
		// if (p.containsKey(addressPropertyKey)) {
		// usvAddress = p.getProperty(addressPropertyKey);
		// }
		//
		// } else {
		// throw new IllegalArgumentException(" Die Datei " + fileLocation
		// + " existiert nicht, bzw. kann nicht gelesen werden.");
		// }
		// } catch (final IOException ex) {
		// // Sollte nie passieren, nachdem wir schon geprueft haben, ob die
		// // Datei existiert und gelesen werden kann.
		// throw new IllegalArgumentException(" Auf die Properties-Datei "
		// + fileLocation + " kann nicht zugegriffen werden.", ex);
		// }

		usvAddress = USVWrapper.getSnmpServer();

		anmeldenAnTraps();
	}

	/**
	 * Wir melden uns mit dieser Methode an sog. SNMP Traps an. Wird ein solcher
	 * Trap empfangen, dann wird {@link USVWrapper#setTrigger()} aufgerufen und
	 * die Betriebsdaten der USV aktualisiert.
	 */
	private void anmeldenAnTraps() {
		try {
			final Address listenAddress = GenericAddress.parse("udp:0.0.0.0/162");

			final TransportMapping transport = new DefaultUdpTransportMapping();
			final Snmp snmp = new Snmp(transport);
			snmp.listen();

			final CommandResponder trapPrinter = new CommandResponder() {
				@Override
				public synchronized void processPdu(final CommandResponderEvent e) {
					final PDU command = e.getPDU();
					if (command != null) {
						log.info(
								"Es wurde eine Trap (Notifikation) empfangen. Es erfolgt eine sofortige Aktualisierung.");
						getUsvWrapper().setTrigger();
					}
				}
			};
			snmp.addNotificationListener(listenAddress, trapPrinter);

		} catch (final Exception ex) {
			log.error("Die Anmeldung für SNMP Traps ist fehlgeschlagen.", ex);
		}
	}

	/**
	 * Die Defaultimplementierung gibt immer "Funktionalität nicht unterstützt"
	 * zurück.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public int setNotstrom(final boolean befehl) {
		return -1;
	}

	/**
	 * Die Defaultimplementierung gibt immer "Funktionalität nicht unterstützt"
	 * zurück.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public int setUsvBetrieb(final boolean befehl) {
		return -1;
	}

	/**
	 * Gibt die Antwort auf die Anfrage nach eienr einzelnen Variable zurück.
	 * Gibt {@code null} zurück, wenn keine Antwort ermittelbar ist, z.&nbsp;B.
	 * bei einem Timeout.
	 */
	private PDU sendRequest(final String oid) {
		final Snmp snmp;
		final CommunityTarget target;

		try {
			final Address targetAddress = GenericAddress.parse(usvAddress);
			final TransportMapping transport = new DefaultUdpTransportMapping();
			snmp = new Snmp(transport);
			target = new CommunityTarget();
			target.setCommunity(new OctetString("public"));
			target.setAddress(targetAddress);
			target.setRetries(2);
			target.setTimeout(timeout);
			target.setVersion(SnmpConstants.version2c);
			snmp.listen();
		} catch (final IOException ex) {
			log.error("Konnte keine Verbindung zur USV aufbauen.", ex);
			return null;
		}

		final PDU pdu = new PDU();
		PDU result = null;
		pdu.add(new VariableBinding(new OID(oid)));
		pdu.setType(PDU.GET);
		try {
			final PDU response = snmp.get(pdu, target).getResponse();
			if (response != null) {
				if (response.size() == 1) {
					result = response;
				} else {

					// Sollte eigentlich nicht passieren
					log.error("Die Antwort von der USV hat nicht die erwartete Anzahl Elemente (1	 Element).",
							response);
				}
			} else {
				log.error("Die USV hat nicht innerhalb des Timeouts geantwortet (OID=" + oid + ").");
			}
		} catch (final IOException ex) {
			log.error("Anfrage konnte nicht an USV gesendet werden: " + pdu, ex);
		} finally {
			try {
				snmp.close();
			} catch (final IOException ex) {
				log.error("Konnte Verbindung zur USV nicht erfolgreich schließen.", ex);
			}
		}

		return result;
	}

}
